@@ -36,6 +36,8 @@
 #include "true.h"
 #endif
 
+#include <ctype.h>
+
 #include "debug.h"
 
 void httpdata_init(struct httpdata *e)
@@ -80,7 +82,15 @@ static const char* mime_file[] =
 	NULL
 };
 static const char* mime_m3u[] = { "audio/mpegurl", "audio/mpeg-url", "audio/x-mpegurl", NULL };
-static const char* mime_pls[]	= { "audio/x-scpls", "audio/scpls", "application/pls", "application/x-scpls", NULL };
+static const char* mime_pls[]	=
+{
+	"audio/x-scpls"
+,	"audio/scpls"
+,	"application/pls"
+,	"application/x-scpls"
+,	"application/pls+xml"
+,	NULL
+};
 static const char** mimes[] = { mime_file, mime_m3u, mime_pls, NULL };
 
 int debunk_mime(const char* mime)
@@ -93,13 +103,17 @@ int debunk_mime(const char* mime)
 	aux = strchr(mime, ';');
 	if(aux != NULL)
 	{
-		fprintf(stderr, "Warning: additional info in content-type ignored (%s)\n", aux+1);
-		/* Just compare up to before the ";". */
+		if(!param.quiet)
+			fprintf(stderr, "Warning: additional info in content-type ignored (%s)\n", aux+1);
+		/* Just compare up to before the ";" */
 		len = aux-mime;
 	}
 	/* Else, compare the whole string -- including the end. */
 	else len = strlen(mime)+1;
 
+	/* Skip trailing whitespace, to ne nice to strange folks. */
+	while(len && isspace(mime[len-1])) --len;
+
 	for(i=0; mimes[i]    != NULL; ++i)
 	for(j=0; mimes[i][j] != NULL; ++j)
 	if(!strncasecmp(mimes[i][j], mime, len)) goto debunk_result;
@@ -146,10 +160,10 @@ static int writestring (int fd, mpg123_string *string)
 	return TRUE;
 }
 
-static size_t readstring (mpg123_string *string, size_t maxlen, FILE *f)
+static size_t readstring (mpg123_string *string, size_t maxlen, int fd)
 {
 	int err;
-	debug2("Attempting readstring on %d for %"SIZE_P" bytes", f ? fileno(f) : 0, (size_p)maxlen);
+	debug2("Attempting readstring on %d for %"SIZE_P" bytes", fd, (size_p)maxlen);
 	string->fill = 0;
 	while(maxlen == 0 || string->fill < maxlen)
 	{
@@ -160,7 +174,7 @@ static size_t readstring (mpg123_string *string, size_t maxlen, FILE *f)
 			string->fill = 0;
 			return 0;
 		}
-		err = read(fileno(f),string->p+string->fill,1);
+		err = read(fd,string->p+string->fill,1);
 		/* Whoa... reading one byte at a time... one could ensure the line break in another way, but more work. */
 		if( err == 1)
 		{
@@ -497,7 +511,6 @@ int http_open(char* url, struct httpdata *hd)
 	int oom  = 0;
 	int relocate, numrelocs = 0;
 	int got_location = FALSE;
-	FILE *myfile = NULL;
 	/*
 		workaround for http://www.global24music.com/rautemusik/files/extreme/isdn.pls
 		this site's apache gives me a relocation to the same place when I give the port in Host request field
@@ -597,15 +610,10 @@ int http_open(char* url, struct httpdata *hd)
 		
 		if(param.verbose > 2) fprintf(stderr, "HTTP request:\n%s\n",request.p);
 		if(!writestring(sock, &request)){ http_failure; }
-		if (!(myfile = fdopen(sock, "rb")))
-		{
-			error1("fdopen: %s", strerror(errno));
-			http_failure;
-		}
 		relocate = FALSE;
 		/* Arbitrary length limit here... */
 #define safe_readstring \
-		readstring(&response, SIZE_MAX/16, myfile); \
+		readstring(&response, SIZE_MAX/16, sock); \
 		if(response.fill > SIZE_MAX/16) /* > because of appended zero. */ \
 		{ \
 			error("HTTP response line exceeds max. length"); \
@@ -673,6 +681,15 @@ int http_open(char* url, struct httpdata *hd)
 				}
 			}
 		} while(response.p[0] != '\r' && response.p[0] != '\n');
+		if(relocate)
+		{
+			close(sock);
+			sock = -1;
+			/* Forget content type, might just relate to a displayed error page,
+			   not the resource being redirected to. */
+			mpg123_free_string(&hd->content_type);
+			mpg123_init_string(&hd->content_type);
+		}
 	} while(relocate && got_location && purl.fill && numrelocs++ < HTTP_MAX_RELOCATIONS);
 	if(relocate)
 	{
